/* Copyright (C) 1997 Free Software Foundation, Inc.

 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * To receive a copy of the GNU General Public License, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA; or view
 * http://www-swiss.ai.mit.edu/~jaffer/GPL.html
 */

/* "!#.c", part of the #! shell-script trampoline for DOS.
   Author: Aubrey Jaffer */

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
				/* Borland C++ 3.1 */
#ifdef __MSDOS__
# include <process.h>
# include <dir.h>
# include <io.h>
#endif
				/* DJGPP 2.1 needs unistd.h */
#ifdef unix
# include <unistd.h>
#endif
#include <stdio.h>

int	script_P(const char *name);
char *	copy_path (const char *name);
char *	dos_find_ext(const char *name,
		     const char *ext,
		     int (*checker)(const char *name));
char *	dos_find_(const char *name,
		  const char *ext,
		  int (*checker)(const char *name));
char *read_script_prog(const char *name);
char *rsp(const char *name);
int main(int argc, char **argv);

char *copy_path(name)
     const char *name;
{
  char *str1 = (char *)malloc((size_t)(strlen(name) + 1));
  if (!str1) return 0L;
  str1[0] = 0;
  strcat(str1, name);
  return str1;
}

# define DEFAULT_PATH "c:\\dos"
# define PATH_DELIMITER ';'
# define ABSOLUTE_FILENAME_P(fname) ((fname[0] == '\\') \
				     || (fname[0] && (fname[1] == ':')))

int script_P(name)
     const char *name;
{
  char *prog = read_script_prog(name);
  if (!prog) return 0;
  free(prog); return !0;
}
char *read_script_prog(name)
     const char *name;
{
  char *sbuf;
  int ind = 0, c;
  FILE *f = fopen(name, "r");
  if (!f) return 0;
  if ('#' != fgetc(f)) goto ret0;
  if ('!' != fgetc(f)) goto ret0;
  while (1) {
    c = fgetc(f);
    switch (c) {
    case EOF: goto ret0;
    case ' ': case '\t': case '\r': case '\n':
      continue;
    default:
      sbuf = (char *)malloc(1);
      if (!sbuf) goto ret0;
      goto startit;
    }
  }
  while (1) {
    c = fgetc(f);
  startit:
    sbuf[ind] = c;
    switch (c) {
    case EOF: case '\r': case '\n':
      goto fret0;
    case ' ': case '\t':
      sbuf[ind] = 0;
      fclose(f);
      if (ind) return sbuf;
      goto fret0;
    default:
      sbuf = (char *)realloc(sbuf, ++ind + 1);
      if (!sbuf) goto ret0;
      continue;
    }
  }
fret0:
  free(sbuf);
ret0:
  fclose(f);
  return 0;
}

#ifndef X_OK
# define X_OK 1
#endif /* ndef X_OK */
int exe_P(name)
     const char *name;
{
  return (!access(name, X_OK));
}

#ifndef MAXPATHLEN
# define MAXPATHLEN 80
#endif /* ndef MAXPATHLEN */

char *dos_find_ext(name, ext, checker)
     const char *name, *ext;
     int (*checker)();
{
  char *paths = "", *suf;
  char nbuf[MAXPATHLEN];
  char tbuf[MAXPATHLEN];
  int ind = 0, chr;

  while ((ind < (MAXPATHLEN-1-4)) && name[ind])
    switch (chr = (name[ind])) {
    case '/': chr = '\\';
    default: nbuf[ind++] = chr;
    }
  nbuf[ind] = 0L;
  suf = strchr(nbuf + ind - 4, '.');
  if ((!suf) || strchr(suf + 1, '.'))
    if (ext) {suf = 0; strcat(nbuf, ext); goto suffixed;}
    else suf = nbuf + ind;
  else {suf = 0; goto suffixed;}
  strcpy(suf, ".com");
  if (checker(nbuf)) goto foundit;
  strcpy(suf, ".exe");
  if (checker(nbuf)) goto foundit;
  strcpy(suf, ".bat");
suffixed:
  if (checker(nbuf)) {
  foundit:
    if (ABSOLUTE_FILENAME_P(nbuf)) return copy_path(nbuf);
    ind = 0;
    tbuf[0] = '.';
    if ((!getcwd(tbuf, MAXPATHLEN-1-strlen(nbuf))) || ('.'==tbuf[0]))
      return copy_path(nbuf);
    while ((ind < (MAXPATHLEN-1-strlen(nbuf))) && tbuf[ind])
      switch (chr = (tbuf[ind])) {
      case '/': chr = '\\';
      default: tbuf[ind++] = chr;
      }
    tbuf[ind] = 0L;
    if (suf) strcpy(suf, "");
    goto trycd;
  }
  else if (suf) strcpy(suf, "");
  if (ABSOLUTE_FILENAME_P(nbuf) || (strchr(nbuf, '\\'))) return 0L;

  if ((!(paths = (char *) getenv("#!PATH"))) &&
      (!(paths = (char *) getenv("PATH"))))
    paths = DEFAULT_PATH;

  while (*paths) {
    ind = 0;
    while (*paths && (PATH_DELIMITER != *paths))
      switch (chr = (*paths++)) {
      case '/': chr = '\\';
      default: tbuf[ind++] = chr;
      }
    tbuf[ind] = 0L;
    if (*paths) paths++;
    if (('.'==tbuf[0]) && !tbuf[1]) continue;
  trycd:
    if ('\\' != tbuf[ind - 1]) strcat(tbuf, "\\");
    strcat(tbuf, nbuf);
    if (suf) {
      suf = tbuf + strlen(tbuf);
      strcpy(suf, ".com");
      if (checker(tbuf)) return copy_path(tbuf);
      strcpy(suf, ".exe");
      if (checker(tbuf)) return copy_path(tbuf);
      strcpy(suf, ".bat");
    }
    if (checker(tbuf)) return copy_path(tbuf);
  }
  return 0L;
}

int main(argc, argv)
     int argc;
     char **argv;
{
  char *name;
  int ind;
tail:
  if (3 > argc) goto errout;
  /* fprintf(stderr, "main %d %s(%s", argc, argv[0], argv[1]);
     for (ind = 2; ind < argc; ind++) fprintf(stderr, ", %s", argv[ind]);
     fprintf(stderr, ")\n"); */
  ind = 0;
  if ('#'==argv[1][0] && '!'==argv[1][1]) argv[1]+=2;
  if (!argv[1][0]) {argc--; argv++;}
  name = argv[(strcmp("\\",argv[2])) ? 2 : 3];
  if (!name) {name = argv[2]; goto nofind;}
  if ((name = dos_find_ext(name, ".bat", script_P)))
    argv[(strcmp("\\",argv[2])) ? 2 : 3] = name;
  else {
    name = argv[(strcmp("\\",argv[2])) ? 2 : 3];
  nofind:
    fprintf(stderr, "File not found - %s%s\n",
	    name, strchr(name,'.') ? "" : ".bat");
    goto erroutq;
  }
  if (strchr(argv[1], '/')) {
    while ((name = strchr(argv[1] , '/'))) *name = '\\';
    name = dos_find_ext(argv[1], 0L, exe_P);
    if (name) goto execit;
    name = strrchr(argv[1], '\\');
    argv[1] = 1+name;
  }
  name = dos_find_ext(argv[1], 0L, exe_P);
execit:
  if (name) argv[1] = name;
  if (stricmp(".bat", argv[1] + strlen(argv[1]) - 4)) {argc--; argv++;}
  else {
    if (script_P(argv[1])) {
      char **nargv = (char **)malloc((argc + 1 + 2) * sizeof(char **));
      int nargc = argc + 2;
      if (!nargv) goto errout;
      nargv[0] = argv[0];
      nargv[1] = read_script_prog(argv[1]);
      nargv[2] = "\\";
      for (ind = 1; argv[ind]; ind++) nargv[ind + 2] = argv[ind];
      nargv[ind + 2] = 0;
      argc = nargc; argv = nargv; goto tail;
    }
    name = (char *)malloc(strlen(argv[1]) + 1 + 2);
    if (!name) goto errout;
    strcpy(name, "/C");
    strcat(name, argv[1]);
    argv[1] = name;
    argv[0] = "command";
  }
  for(ind = 0; argv[ind]; ind++) {
    name = copy_path(argv[ind]);
    if (!name) goto errout;
    argv[ind] = name;
  }
  execvp(argv[0], &argv[0]);
errout:  
  fprintf(stderr, "#! %s", argv[0]);
  for (ind = 1; ind < argc; ind++) fprintf(stderr, ", %s", argv[ind]);
  fprintf(stderr, "\n");
  if (3 > argc) fprintf(stderr, "required parameter missing\n");
  else
    erroutq:
  perror("#! failed");
  return EXIT_FAILURE;
}
